import jax
import jax.numpy as jnp
from jaxfmm.transf import M2M, M2L, L2L
from jaxfmm.basis import eval_regular_basis, eval_regular_basis_grad, inv_mpl_idx
from jaxfmm.hierarchy import handle_padding
from functools import partial

__all__ = ["eval_potential", "eval_potential_direct"]

@partial(jax.jit, static_argnames=['p', 'mem_limit'])
def get_initial_mpls(padded_pts, padded_chrgs, boxcenters, p, mem_limit = jnp.inf):
    r"""
    Get initial multipole expansions for each box on the highest level.
    """
    mem = padded_chrgs.size * (p+1)**2 * padded_chrgs.dtype.itemsize
    batch_size = int((mem_limit/mem) * boxcenters.shape[0]) + 1 if mem > mem_limit else boxcenters.shape[0]

    def impl_body(args):
        padded_pts_loc, padded_chrgs_loc, boxcenters_loc = args
        dist = padded_pts_loc - boxcenters_loc
        return (eval_regular_basis(dist,p) * padded_chrgs_loc).sum(axis=0)
    return jax.lax.map(impl_body, [padded_pts, padded_chrgs[...,None], boxcenters[:,None,:]], batch_size=batch_size)

@partial(jax.jit, static_argnames=['n_l', 'mem_limit'])
def go_down(coeff, boxcenters, unqs, invs, n_l, mem_limit = jnp.inf):
    r"""
    Using multipole-to-multipole transformation, descend the hierarchy.
    """
    mpls = [coeff]
    for i in range(n_l,0,-1):
        mpls.append(M2M(mpls[-1],boxcenters[i],boxcenters[i-1], unqs, invs, mem_limit))
    return mpls

@partial(jax.jit, static_argnames=['lvl_info', 'mem_limit'])
def go_up(mpls, locs, boxcenters, eval_boxcenters, mpl_cnct, unqs, invs, lvl_info, img_cnct, mem_limit = jnp.inf):
    r"""
    Using multipole-to-local and local-to-local transformation, ascend the hierarchy.
    """
    l_eval, _ = lvl_info[0]
    for j in range(0, l_eval):   # shift the PBC local expansion to the bottom level in the connectivity
        locs = L2L(locs, eval_boxcenters[j], eval_boxcenters[j+1], unqs, invs, mem_limit)

    l_src_max = len(mpls) - 1
    for i in range(len(lvl_info) - 1):  # for every level pair in the connectivity
        l_eval, l_src = lvl_info[i]
        locs += M2L(boxcenters[l_src], eval_boxcenters[l_eval], mpl_cnct[i], unqs, invs, mpls[l_src_max-l_src], img_cnct, mem_limit)
        if(i < (len(lvl_info) - 2)):    # skip direct connectivity at the end
            for j in range(l_eval, lvl_info[i+1][0]):   # move on to the next eval level, shift multiple times if necessary
               locs = L2L(locs,eval_boxcenters[j],eval_boxcenters[j+1], unqs, invs, mem_limit)
    return locs

@partial(jax.jit, static_argnames=['p', 'field', 'mem_limit'])
def eval_local(locs, padded_eval_pts, rev_idcs, boxcenters, p, field = False, mem_limit = jnp.inf):
    r"""
    Evaluate local expansions on the highest level.
    """
    mem = rev_idcs.size * (p+1)**2 * (1+2*field) * locs.dtype.itemsize
    batch_size = int((mem_limit/mem) * locs.shape[0]) + 1 if mem > mem_limit else locs.shape[0]
    def evloc_body(args):
        pts_loc, box_loc, loc_loc = args
        if(field):
            reg = eval_regular_basis_grad(pts_loc - box_loc, p)
            loc_loc = loc_loc[...,None]  # need additional newaxis for vector components
        else:
            reg = eval_regular_basis(pts_loc - box_loc,p)
        ms, _ = inv_mpl_idx(jnp.arange((p+1)**2))
        prefac = ((2-(ms==0))*(1-2*(ms<0)))[None,None,:]
        if(field):
            prefac = prefac[...,None]
        padded_res = (loc_loc * reg * prefac).sum(axis=2)
        return padded_res
    padded_res = jax.lax.map(evloc_body, [padded_eval_pts, boxcenters[:,None], locs[:,None]], batch_size=batch_size)
    if(field):
        return -padded_res.reshape((-1,3))[rev_idcs] / (4*jnp.pi)
    else:
        return padded_res.flatten()[rev_idcs] / (4*jnp.pi)

@partial(jax.jit, static_argnames=['field', 'mem_limit'])
def eval_direct(padded_pts, padded_chrgs, padded_eval_pts, rev_idcs, dir_cnct, img_cnct = jnp.array([[]]), field = False, mem_limit = jnp.inf):
    r"""
    Evaluate the near-field potential directly (P2P).
    """
    if(dir_cnct.shape[1] == 0):   # nothing to compute
        return 0
    nboxs_src = padded_pts.shape[0]
    distsvec = padded_eval_pts[:,None,None,:,:]
    mem = dir_cnct.shape[0] * dir_cnct.shape[1] * padded_eval_pts.shape[1] * padded_pts.dtype.itemsize * 3  # TODO: why do we remove padded_pts.shape[1]? replace with a sqrt of the product with padded_pts.shape[1]?
    batch_size = int((mem_limit/mem) * padded_eval_pts.shape[0]) + 1 if mem > mem_limit else padded_eval_pts.shape[0]

    def dir_batched(args):
        partner, distsvec_loc = args
        if(img_cnct.shape[1] != 0): # PBC enabled
            distsvec_loc -= img_cnct.at[partner//nboxs_src,None,None,:].get(mode="fill",fill_value=jnp.inf) # image offset
            partner %= nboxs_src    # local image position
        distsvec_loc -= padded_pts[partner,:,None,:]
        distsnorm = jnp.linalg.norm(distsvec_loc,axis=-1)
        distsnorm = 1/jnp.where(distsnorm==0,jnp.inf,distsnorm)
        chrgs = padded_chrgs.at[partner].get(mode="fill",fill_value=0.0)
        if(field):
            pot = ((chrgs[...,None]*(distsnorm**3))[...,None] * distsvec_loc).sum(axis=(0,1))
        else:
            pot = (distsnorm*chrgs[...,None]).sum(axis=(0,1))
        return pot
    pot = jax.lax.map(dir_batched, [dir_cnct, distsvec], batch_size = batch_size)
    if(field):
        return pot.reshape((-1,3))[rev_idcs]/(4*jnp.pi)
    else:
        return pot.flatten()[rev_idcs]/(4*jnp.pi)

@partial(jax.jit, static_argnames=['field', 'lvl_info', 'mem_limit'])
def eval_potential(chrgs, pts, eval_pts, idcs, boxcenters, eval_boxcenters, mpl_cnct, dir_cnct, unqs, invs, lvl_info, img_cnct, PBC_op, field = False, mem_limit = jnp.inf, **kwargs):
    r"""
    Evaluate the potential/field via FMM. Requires an array of charge values and the hierarchy information generated by gen_hierarchy.
    """
    if(len(lvl_info) == 1): # only direct interactions - compute potential directly
        return eval_potential_direct(pts,chrgs,eval_pts,field)
    p = invs[-1].shape[0] - 1
    pad_arr = handle_padding(pts, chrgs, eval_pts, idcs)
    coeff = get_initial_mpls(pad_arr[0], pad_arr[1], boxcenters[lvl_info[-2][1]], p, mem_limit)
    mpls = go_down(coeff, boxcenters, unqs, invs, lvl_info[-2][1], mem_limit)
    locs = (PBC_op@mpls[-1][0])[None,:]    # PBC_op is 0 for open boundary conditions
    locs = go_up(mpls, locs, boxcenters, eval_boxcenters, mpl_cnct, unqs, invs, lvl_info, img_cnct, mem_limit)
    return eval_local(locs, pad_arr[2], idcs[1][1], eval_boxcenters[lvl_info[-2][0]], p, field, mem_limit) + \
           eval_direct(pad_arr[3], pad_arr[4], pad_arr[5], idcs[3][1], dir_cnct, img_cnct, field, mem_limit)

@partial(jax.jit, static_argnames=['field'])
def eval_potential_direct(pts, chrgs, eval_pts = None, field = False):
    r"""
    Evaluate the potential directly via pairwise sums.

    :param pts: Array containing point positions.
    :type padded_pts: jnp.array
    :param chrgs: Array containing point charges.
    :type chrgs: jnp.array
    :param eval_pts: Array containing points to evaluate the potential at. Defaults to pts.
    :type eval_pts: jnp.array, optional
    :param field: Optionally evaluate the field (negative gradient) instead of the potential.
    :type field: bool, optional

    :return: Electrostatic potential (or field) of the points and corresponding charges.
    :rtype: jnp.array
    """
    if(eval_pts is None):
        eval_pts = pts
    res = jnp.zeros(eval_pts.shape[:field+1])
    def eval_direct_body(i, val):
        distsvec = pts[:,:] - eval_pts[i,None,:]
        inv_dists = jnp.linalg.norm(distsvec,axis=-1)
        inv_dists = 1/jnp.where(inv_dists==0,jnp.inf,inv_dists) # take out self-interaction
        if(field):
            val = val.at[i].set(-((chrgs * inv_dists**3)[:,None] * distsvec).sum(axis=0))
        else:
            val = val.at[i].set((chrgs * inv_dists).sum())
        return val
    return jax.lax.fori_loop(0,eval_pts.shape[0],eval_direct_body,res)/(4*jnp.pi)


import pytest
import jaxfmm.fmm as fmm
import jax.numpy as jnp
from jax import random as rng

@pytest.mark.parametrize('unitcube', [[2**15, {}], [2**15, {"p": 4, "eval_pts": rng.uniform(rng.key(7),(2**12,3),minval=0.25,maxval=0.75)}]], indirect=True)
def test_potential(unitcube):
    chrgs, hier = unitcube
    res = fmm.eval_potential(chrgs, **hier)
    res_dir = fmm.eval_potential_direct(hier["pts"], chrgs, hier["eval_pts"])
    L2err = jnp.linalg.norm(res - res_dir)/jnp.linalg.norm(res_dir)
    assert L2err < 6e-3

@pytest.mark.parametrize('unitcube', [[2**15, {}]], indirect=True)
def test_field(unitcube):
    chrgs, hier = unitcube
    res = fmm.eval_potential(chrgs,**hier, field=True)
    res_dir = fmm.eval_potential_direct(hier["pts"],chrgs, field=True)
    L2err = jnp.linalg.norm(res-res_dir)/jnp.linalg.norm(res_dir)
    assert L2err < 3e-3


import jax.numpy as jnp
from jaxfmm import *
from jax import random
import pytest

@pytest.mark.parametrize("periodic_axes", [(0,),(1,),(2,),(0,1),(0,2),(1,2),(0,1,2)])
def test_pbc(periodic_axes):    # TODO: use unitcube fixture, improve test
    ### generate system
    N = 128
    pbc_lvls = 1
    key = random.key(768)
    pts = random.uniform(key,(N,3),minval=-1,maxval=1)
    chrgs = random.uniform(key,N,minval=-1,maxval=1)

    ### balance charges
    possum = chrgs[chrgs>0].sum()
    negsum = jnp.abs(chrgs[chrgs<0].sum())
    chrgs = chrgs.at[chrgs>0].multiply(negsum/possum)

    ### compute periodic potential
    tree_info = gen_hierarchy(pts, N_max=64, periodic_axes=periodic_axes, pbc_lvls=pbc_lvls, debug_info=True, L0_boxcen=jnp.array([0.0,0.0,0.0]),L0_boxlen=jnp.array([2.0,2.0,2.0]), pbc_no_monopole=False)
    pot_FMM = eval_potential(chrgs, **tree_info)

    ### dupe the points - NOTE: only works for cubic level zero cells
    dupes = [(tree_info["pbc_ws"][0][0].shape[0])**((pbc_lvls+2.0)/len(periodic_axes)) if i in periodic_axes else 0 for i in range(3)]
    tmp = jnp.mgrid[-(dupes[0]//2):(dupes[0]//2)+1,-(dupes[1]//2):(dupes[1]//2)+1,-(dupes[2]//2):(dupes[2]//2)+1].T.reshape((-1,3)) * tree_info["boxlens"][0][0]
    pts_nonper = (pts[None,:,:] + tmp[:,None,:]).reshape((-1,3))
    chrgs_nonper = jnp.tile(chrgs, (tmp.shape[0],))

    ### compute analytic potential
    pot_dir = eval_potential_direct(pts_nonper,chrgs_nonper,pts)
    err = jnp.linalg.norm(pot_dir - pot_FMM)/jnp.linalg.norm(pot_dir)

    assert err < 3e-2


import pytest
from jax import random
from jaxfmm.hierarchy import gen_hierarchy

@pytest.fixture(scope="module")
def unitcube(request):
    if(len(request.param)==2):
        seed = 23556
        N, kwargs = request.param
    else:
        N, seed, kwargs = request.param
    pts = random.uniform(random.key(seed),(N,3),minval=-0.5,maxval=0.5)
    chrgs = random.uniform(random.key(seed+1),N,minval=-1,maxval=1)
    hier = gen_hierarchy(pts,**kwargs)
    return chrgs, hier


import jax
import jax.numpy as jnp
from jaxfmm.rotation import get_polar_rot_coeff, rot_azimuth, rot_polar, cart_to_sph
from jax.scipy.special import factorial as fac
from jaxfmm.basis import mpl_idx, inv_mpl_idx
from functools import partial

@partial(jax.jit, static_argnames=["mem_limit"])
def M2M(coeffs, oldboxdims, newboxdims, unqs, invs, mem_limit = jnp.inf):
    p = invs[-1].shape[0]-1
    n_chi = oldboxdims.shape[0]//newboxdims.shape[0]
    coeffs = coeffs.reshape((coeffs.shape[0]//n_chi,n_chi,coeffs.shape[1]))
    oldboxdims = oldboxdims.reshape((-1,n_chi,3))
    mem = unqs.shape[1] * oldboxdims.shape[0] * n_chi * coeffs.dtype.itemsize
    batch_size = int((mem_limit/mem) * coeffs.shape[0]) + 1 if mem > mem_limit else coeffs.shape[0]

    def M2M_batched(args):
        coeffs, oldboxdims, newboxdims = args
        sph = cart_to_sph(oldboxdims - newboxdims)[None,...]
        ### get rotation matrix components all at once
        basemat = get_polar_rot_coeff(p,unqs,sph[...,1])
        ### rotate
        coeffs = rot_azimuth(coeffs, sph[...,2])
        coeffs = rot_polar(basemat,invs,coeffs)
        ### shift in the rotated system
        new_coeffs = coeffs.copy()
        for j in range(1,p+1):
            fact = (sph[...,0]**j) / fac(j)
            ms, ns = inv_mpl_idx(jnp.arange((p-j+1)**2))
            idxs = ms + ns + (ns+j)**2 + j
            new_coeffs = new_coeffs.at[...,idxs].add(fact[...,None] * coeffs[...,:mpl_idx(p-j,p-j)+1])
        ### rotate back
        coeffs = rot_polar(basemat,invs,new_coeffs,True)
        coeffs = rot_azimuth(coeffs, sph[...,2],True)
        return coeffs.sum(axis=(0,1))   # TODO: remove one sum?
    return jax.lax.map(M2M_batched,[coeffs,oldboxdims,newboxdims[:,None,:]],batch_size = batch_size)

@partial(jax.jit, static_argnames=["mem_limit"])
def L2L(coeffs_glob, oldboxdims, newboxdims, unqs, invs, mem_limit = jnp.inf):
    p = invs[-1].shape[0]-1
    n_chi = newboxdims.shape[0]//oldboxdims.shape[0]
    coeffs_glob = coeffs_glob[:,None,:]
    newboxdims = newboxdims.reshape((-1,n_chi,3))
    mem = unqs.shape[1] * newboxdims.shape[0] * n_chi * coeffs_glob.dtype.itemsize
    batch_size = int((mem_limit/mem) * coeffs_glob.shape[0]) + 1 if mem > mem_limit else coeffs_glob.shape[0]

    def L2L_batched(args):
        coeffs, oldboxdims, newboxdims = args
        sph = cart_to_sph(newboxdims - oldboxdims)[None,...]    # TODO: remove the None here?
        ### get rotation matrix components all at once
        basemat = get_polar_rot_coeff(p,unqs,sph[...,1])
        ### rotate
        coeffs = rot_azimuth(coeffs, sph[...,2],False,True)
        coeffs = rot_polar(basemat,invs,coeffs,False,True)
        ### shift in the rotated system
        new_coeffs = coeffs.copy()
        for j in range(1,p+1):
            fact = (sph[...,0]**j) / fac(j)
            ms, ns = inv_mpl_idx(jnp.arange((p-j+1)**2))
            idxs = ms + ns + (ns+j)**2 + j
            new_coeffs = new_coeffs.at[...,:mpl_idx(p-j,p-j)+1].add(fact[...,None] * coeffs[...,idxs])
        ### rotate back
        coeffs = rot_polar(basemat,invs,new_coeffs,True,True)
        coeffs = rot_azimuth(coeffs, sph[...,2],True,True)
        return coeffs
    coeffs = jax.lax.map(L2L_batched,[coeffs_glob,oldboxdims[:,None,:],newboxdims],batch_size = batch_size)
    return coeffs.reshape((-1,coeffs.shape[-1]))

@partial(jax.jit, static_argnames=['mem_limit'])
def M2L(boxcenters, eval_boxcenters, mpl_cnct, unqs, invs, coeffs, img_cnct = jnp.array([[]]), mem_limit = jnp.inf):
    p = invs[-1].shape[0]-1
    n_boxs = boxcenters.shape[0]
    eval_boxcenters = eval_boxcenters[:,None,:]
    mem = unqs.shape[1] * mpl_cnct.shape[0] * mpl_cnct.shape[1] * coeffs.dtype.itemsize / 2 # TODO: why the factor 2?
    batch_size = int((mem_limit/mem) * mpl_cnct.shape[0]) + 1 if mem > mem_limit else mpl_cnct.shape[0]

    def M2L_batched(argss):
        eval_boxcenters, mpl_cnct = argss
        if(img_cnct.shape[1] > 0):  # both pbc images and real boxes, shift appropriately
            ids = mpl_cnct%n_boxs
            sph = cart_to_sph(eval_boxcenters - boxcenters[ids] - img_cnct.at[mpl_cnct//n_boxs].get(mode="fill",fill_value=1.23456789))   # TODO: see below
            ids += ((mpl_cnct//n_boxs)//img_cnct.shape[0])*n_boxs   # edit ids such that the padding is out of bounds
        else:   # open boundary conditions
            ids = mpl_cnct
            sph = cart_to_sph(eval_boxcenters - boxcenters.at[ids].get(mode="fill",fill_value=1.23456789))    # TODO: replace this workaround with a nan_to_num call?
        coeffs_loc = coeffs.at[ids].get(mode="fill",fill_value=0)   # the fill value ensures that padding is ignored
        sph = sph[None,...]
        ### get rotation matrix components all at once
        basemat = get_polar_rot_coeff(p,unqs,sph[...,1])
        ### rotate
        coeffs_loc = rot_azimuth(coeffs_loc, sph[...,2])
        coeffs_loc = rot_polar(basemat,invs,coeffs_loc)
        ### carry out M2L transformation in the rotated system
        for j in range(p+1):
            ns = jnp.arange(j,p+1)
            pws = ns[:,None] + ns[None,:]
            fact = (-1)**(ns[:,None]+j) * fac(pws)[None,None,...] / ((sph[...,0,None,None])**((pws + 1)[None,None,...])) # NOTE: the factor (-1) in front probably originates from the sign of the shift vector...
            ### real update
            idxs = mpl_idx(j,ns)
            coeffs_loc = coeffs_loc.at[...,idxs].set((fact * coeffs_loc[...,None,idxs]).sum(axis=-1))
            if(j>0):    # imag update
                idxs = mpl_idx(-j,ns)
                coeffs_loc = coeffs_loc.at[...,idxs].set(-(fact * coeffs_loc[...,None,idxs]).sum(axis=-1))  # imag update - note the inverse sign!
        ### rotate back
        coeffs_loc = rot_polar(basemat,invs,coeffs_loc,True,True)
        coeffs_loc = rot_azimuth(coeffs_loc, sph[...,2],True,True)
        return coeffs_loc.sum(axis=(0,1))   # TODO: remove one sum?
    return jax.lax.map(M2L_batched,[eval_boxcenters, mpl_cnct],batch_size = batch_size)


import pytest
import jax.numpy as jnp
import jaxfmm.fmm as fmm
import jaxfmm.debug_helpers as dbg
from jaxfmm.basis import mpl_idx
from jaxfmm.hierarchy import gen_hierarchy

@pytest.mark.parametrize('unitcube', [[128, {}]], indirect=True)
def test_mpl(unitcube):
    chrgs, hier = unitcube
    mpls = fmm.get_initial_mpls(hier["pts"][None,...],chrgs[None,...],hier["boxcenters"][0],hier["p"])
    eval_pts = hier["pts"] + jnp.array([1.0,-2.0,1.14])

    pot_mpl = dbg.eval_mpls(mpls,eval_pts,hier["boxcenters"][0][0],hier["p"])
    pot_dir = fmm.eval_potential_direct(hier["pts"],chrgs,eval_pts)
    assert jnp.allclose(pot_dir,pot_mpl,rtol=1e-3,atol=1e-3)

@pytest.mark.parametrize('unitcube', [[128, {}]], indirect=True)
def test_loc(unitcube):
    chrgs, hier = unitcube
    shift = jnp.array([1.0,-2.0,1.44])
    locs = dbg.get_locs(hier["pts"],chrgs,hier["boxcenters"][0]+shift,hier["p"])
    eval_pts = hier["pts"] + shift

    pot_loc = fmm.eval_local(locs,eval_pts[None,...],jnp.arange(eval_pts.shape[0]),hier["boxcenters"][0]+shift,hier["p"])
    pot_dir = fmm.eval_potential_direct(hier["pts"],chrgs,eval_pts)
    assert jnp.allclose(pot_dir,pot_loc,rtol=1e-3,atol=1e-3)

def test_mpl_coeffs():  # TODO: also check for correct normalization
    for n in range(9):
        for m in range(-n,n+1):
            pts, chrgs = dbg.gen_multipole_dist(m,n,eps=10.0)    # special point charge distribution corresponding to multipole moments - set eps large to minimize error
            tree_info = gen_hierarchy(pts)
            coeff = fmm.get_initial_mpls(pts[None,...],chrgs[None,...],tree_info["boxcenters"][0],n)[0,:]
            test = jnp.where(jnp.abs(coeff)>1e-5)[0]   # only the (m,n) coefficient should be nonzero
            assert test.shape[0] == 1
            assert test[0] == mpl_idx(m,n)


import jax
import jax.numpy as jnp
from functools import partial
from math import log2, ceil
from jaxfmm.pbc import gen_img_connectivity, gen_pbc_op
from jaxfmm.rotation import precompute_combs

__all__ = ["gen_hierarchy"]

def get_max_l(N_tot, N_max, s = 3):
    r"""
    Compute number of levels in the hierarchy.
    """
    max_l = int(ceil(log2(N_tot/N_max)/s))
    return 0 if max_l < 0 else max_l

@partial(jax.jit, static_argnames = ["max_l", "s"])
def balanced_tree(pts, max_l, s = 3):
    r"""
    Generate a balanced 2^n-tree hierarchy.
    """
    n_chi = 2**s
    idcs = jnp.arange(pts.shape[0])[None,:]

    for l in range(max_l*s):    # carry out s splits on max_l levels in total (we cannot make this a for_i loop as the shape constantly changes)
        splitpos = idcs.shape[1]//2   # split position in the middle
        needpad = idcs.shape[1]%2     # modulo tells us if padding must be inserted

        pts_sorted = pts.at[idcs].get(mode="fill",fill_value=-jnp.nan**2)  # padded values get converted to NaNs - NOTE: due to the argpartition implementation, we need to make sure that all the NaNs have the same sign
        axis_to_split = jnp.argmax(jnp.nanmax(pts_sorted,axis=1) - jnp.nanmin(pts_sorted,axis=1),axis=1)    # nanmax and -min to ignore NaNs
        idcs = idcs[jnp.arange(idcs.shape[0])[:,None],jnp.argpartition(pts_sorted[jnp.arange(axis_to_split.shape[0]),:,axis_to_split],splitpos,axis=1)] # splitting, the NaNs introduced below get transported to the beginning

        # padding so the next array has the correct shape
        idcs = jax.lax.pad(idcs,pts.shape[0],[(0,0,0),(0,needpad,0)])   # pad at the end with out of range values
        idcs = idcs.reshape((-1,idcs.shape[1]//2))                      # now that we padded, we can safely reshape this
    
    idcs = jnp.sort(idcs,axis=1)    # sorting is good for locality but might be overkill TODO: swap only first and last positions instead of full sort?
    rev_idcs = jnp.argsort(idcs.flatten())[:pts.shape[0]]     # reverse sorting indices, to undo the sorting
    pts_sorted = pts.at[idcs].get(mode="fill",fill_value=jnp.nan)  
    boxcenters, boxlens = [jnp.zeros((n_chi**l,3)) for l in range(max_l+1)], [jnp.zeros((n_chi**l,3)) for l in range(max_l+1)]

    for l in range(max_l,-1,-1):
        minc, maxc = jnp.nanmin(pts_sorted,axis=1), jnp.nanmax(pts_sorted,axis=1)    # nanmax and nanmin to ignore NaNs
        boxlens[l] = maxc - minc    # in principle we only need to save the norm of this, but it is nice to have for visualizations
        boxcenters[l] = minc + boxlens[l]/2
        if(l > 0):
            pts_sorted = pts_sorted.reshape((-1,pts_sorted.shape[1]*n_chi,3))
    return idcs, rev_idcs, boxcenters, boxlens

def reduce_max_lvl(pts, idcs, s, old_max_l, new_max_l):
    r"""
    Reduce max level of a balanced tree after it has already been created.
    """
    ldiff = old_max_l - new_max_l
    idcs = idcs.reshape((idcs.shape[0]//(2**(s*ldiff)),-1))
    trim = idcs.shape[1] - (idcs.size - pts.shape[0])//idcs.shape[0]
    idcs = jnp.sort(idcs,axis=-1)[:,:trim] # we know exactly how much padding is accumulated!
    rev_idcs = jnp.argsort(idcs.flatten())[:pts.shape[0]] # also fix the reverse sorting
    return idcs, rev_idcs

def gen_connectivity(boxcenters, boxlens, eval_boxcenters, eval_boxlens, theta = 0.75, s = 3, no_cross_level = False, periodic_axes = ()): # TODO: is there a way to make this JIT compilable?
    r"""
    Compute connectivity information for a given hierarchy.
    """
    n_l, eval_n_l = len(boxcenters), len(eval_boxcenters)  # number of levels of the source and eval hierarchy
    n_chi = 2**s     # number of child boxes per split
    periodic = len(periodic_axes) > 0
    if(periodic):
        img_cnct = gen_img_connectivity(boxlens[0], theta, periodic_axes)[0]
    else:
        img_cnct = jnp.array([[]])

    n_img = img_cnct.shape[0]
    l_src, l_eval = 0, 0
    mpl_cnct, lvl_info = [], []
    non_wellseps = jnp.arange(n_img)[None,:]
    keepgoing = True
    while(keepgoing):
        evalend, srcend = (l_eval >= (eval_n_l-1)), (l_src >= (n_l-1))
        keepgoing = not (evalend and srcend)    # we keep going until we reach the final level in both hierarchies
        nbox, nbox_eval = boxcenters[l_src].shape[0], eval_boxcenters[l_eval].shape[0]

        if(periodic):
            rhs = boxcenters[l_src][non_wellseps%nbox] + img_cnct.at[non_wellseps//nbox].get(mode="fill",fill_value=jnp.nan)
        else:
            rhs = boxcenters[l_src].at[non_wellseps].get(mode="fill",fill_value=jnp.nan)
        d = jnp.linalg.norm(eval_boxcenters[l_eval][:,None,:] - rhs,axis=-1)  # distances between boxcenters

        r1 = jnp.linalg.norm(eval_boxlens[l_eval][:,None,:],axis=-1)/2  # eval box radii
        r2 = jnp.linalg.norm(boxlens[l_src][non_wellseps%nbox if periodic else non_wellseps],axis=-1)/2 # source box radii
        tmp = r1 > r2
        R = jnp.where(tmp,r1,r2)     # R = max(r1,r2)
        r = jnp.where(tmp,r2,r1)     # r = min(r1,r2)

        ws = (R + theta*r <= theta*d)     # well-separatedness criterion, NaNs always return False
        non_ws = (R + theta*r > theta*d)  # need to check again to exclude NaNs
        ws_nums = ws.sum(axis=1)          # number of well-separated boxes for each box
        non_ws_nums = non_ws.sum(axis=1)  # number of non-well-separated boxes for each box

        to_pad_ws = jnp.max(ws_nums) - ws_nums    # how much padding per box must be inserted
        ws_padding = jnp.repeat(jnp.cumsum(ws_nums),to_pad_ws)   # the correct indices for the insert below
        cnct_info = jnp.insert(non_wellseps[ws],ws_padding,n_img*nbox).reshape(nbox_eval,-1) # padded ws information
        if(cnct_info.shape[1] > 0):       # do not save empty cnct information
            mpl_cnct.append(cnct_info)
            lvl_info.append((l_eval, l_src))

        to_pad_non_ws = jnp.max(non_ws_nums) - non_ws_nums    # how much padding per box must be inserted for non-well-separated boxes
        non_wellsep_padding = jnp.repeat(jnp.cumsum(non_ws_nums),to_pad_non_ws)   # the correct indices for the insert below
        non_wellseps = jnp.insert(non_wellseps[non_ws],non_wellsep_padding,n_img*nbox).reshape((nbox_eval,-1))    # we overwrite the old values here
        if(non_wellseps.shape[1] == 0):   # we are already done, quit out
            break

        r1mean, r2mean = jnp.mean(r1), jnp.mean(r2)
        if((not srcend) and ((r1mean < 1.05 * r2mean) or evalend or no_cross_level)): # TODO: play around with the first condition and factors therein
            l_src += 1          # src_boxes >= eval_boxes -> increase src lvl
            non_wellseps = jnp.repeat(non_wellseps,n_chi,axis=1)*n_chi + jnp.tile(jnp.arange(n_chi),non_wellseps.shape[1])   # indices change
        if((not evalend) and ((r1mean >= 1.05 * r2mean) or srcend or no_cross_level)):
            l_eval += 1             # eval boxes > src boxes -> increase eval lvl
            non_wellseps = jnp.repeat(non_wellseps,n_chi,axis=0)    # size of non-wellseps grows

    lvl_info.append((l_eval, l_src))
    return mpl_cnct, non_wellseps, tuple(lvl_info), img_cnct

def gen_hierarchy(pts, eval_pts = None, N_max = 128, theta = 0.77, s = 3, p = 3, periodic_axes = (), pbc_lvls = 5, pbc_no_monopole = True, L0_boxcen = None, L0_boxlen = None, no_cross_level = False, debug_info = False, mem_limit = None):
    r"""
    Generate the balanced tree and connectivity for the FMM.

    :param pts: Array of shape (N,3) containing the positions of N point charges.
    :type pts: jnp.array
    :param eval_pts, optional: Array of shape (N_eval,3) containing the positions of N evaluation points. Defaults to None, where eval_pts = pts.
    :type eval_pts: jnp.array
    :param N_max: Maximum allowed number of point charges per box.
    :type N_max: int, optional
    :param theta: Well-separatedness parameter, determines accuracy.
    :type theta: float, optional
    :param s: How many splits per level and box are been performed. Each box has 2^s children.
    :type s: int, optional
    :param p: Maximum expansion order.
    :type p: int, optional
    :param periodic_axes: Tuple indicating which dimensions (0, 1 and/or 2) are periodic.
    :type periodic_axes: tuple
    :param pbc_lvls: How many virtual levels are introduced in the hierarchy. If pbc_lvls < 0, only the non-well-separated boxes on level zero will be added.
    :type pbc_lvls: int
    :param pbc_no_monopole: Disable the contribution from distant monopoles for PBC. True by default, to negate possible errors introduced by almost zero total charge.
    :type pbc_no_monopole: bool
    :param L0_boxcen: Center of the source box at level zero, useful for periodic boundary conditions. The default (None) will generate the smallest possible axis-aligned box.
    :type L0_boxcen: jnp.array, optional
    :param L0_boxlen: Sidelengths of the source box at level zero, useful for periodic boundary conditions. The default (None) will generate the smallest possible axis-aligned box.
    :type L0_boxlen: jnp.array, optional
    :param no_cross_level: For pts = eval_pts, disable cross-level comparisons.
    :type no_cross_level: bool, optional
    :param debug_info: Whether to include further information in the hierarchy (boxlengths, PBC shifts).
    :type debug_info: bool, optional
    :param mem_limit: Memory limit (in Bytes) for the potential evaluation. The default (None) sets mem_limit=1 on CPUs and mem_limit=jnp.inf on GPUs. For GPUs, it is recommended to set this to the L2-Cache size of the GPU model.
    :type mem_limit: float, optional

    :return: Dictionary containing full hierarchy information.
    :rtype: dict
    """
    if(eval_pts is None):
        eval_pts = pts
    periodic = len(periodic_axes) > 0
    sz_eps = 1.01   # safety factor for checking if coordinates are inside level zero boxes
    
    max_l = get_max_l(pts.shape[0], N_max, s)
    eval_max_l = get_max_l(eval_pts.shape[0], N_max, s)
    idcs, rev_idcs, boxcenters, boxlens = balanced_tree(pts, max_l, s)

    old_boxcen, old_boxlen = boxcenters[0].copy(), boxlens[0].copy()
    if(L0_boxcen is not None):  # overwrite with specified value
        boxcenters[0] = L0_boxcen[None,:]
    if(L0_boxlen is not None):  # overwrite with specified value
        boxlens[0] = L0_boxlen[None,:]
    if(jnp.any((old_boxcen[0] - old_boxlen[0]/2 - boxcenters[0]) < -sz_eps*boxlens[0]/2) or jnp.any((old_boxcen[0] + old_boxlen[0]/2 - boxcenters[0]) > sz_eps*boxlens[0]/2)):
        raise ValueError("Level zero box does not contain all source points.")

    if(pts is eval_pts):
        eval_idcs, eval_boxcenters, eval_boxlens = idcs, boxcenters, boxlens
    else:
        no_cross_level = False  # we must deal with cross-level contributions
        eval_idcs, _, eval_boxcenters, eval_boxlens = balanced_tree(eval_pts, eval_max_l, s)

        if(periodic):
            if(jnp.any((eval_boxcenters[0] - eval_boxlens[0]/2 - boxcenters[0]) < -sz_eps*boxlens[0]/2) or jnp.any((eval_boxcenters[0] + eval_boxlens[0]/2 - boxcenters[0]) > sz_eps*boxlens[0]/2)):
                raise ValueError("All evaluation points must be contained in the level zero source box for PBC.")
            eval_boxcenters[0], eval_boxlens[0] = boxcenters[0], boxlens[0]  # for PBC, we need identical level zero boxes

    mpl_cnct, dir_cnct, lvl_info, img_cnct = gen_connectivity(boxcenters, boxlens, eval_boxcenters, eval_boxlens, theta, s, no_cross_level, periodic_axes)

    ### we adapt the sorting idcs to the "true" max levels - the if statements prevent unnecessary work
    only_dir = (len(lvl_info) == 1)
    idcs_list = [(idcs, rev_idcs) if only_dir else reduce_max_lvl(pts,idcs,s,max_l,lvl_info[-2][1])] # src padding at max mpl level

    if(only_dir or (pts is eval_pts and lvl_info[-2][0]==lvl_info[-2][1])):   # eval padding at max mpl level
        idcs_list.append(idcs_list[0])
    else:
        idcs_list.append(reduce_max_lvl(eval_pts,eval_idcs,s,eval_max_l,lvl_info[-2][0]))

    if(only_dir or (lvl_info[-2][1]==lvl_info[-1][1])):                       # src padding at max dir level
        idcs_list.append(idcs_list[0])
    else:
        idcs_list.append(reduce_max_lvl(pts,idcs,s,max_l,lvl_info[-1][1]))

    if(only_dir or (lvl_info[-2][0]==lvl_info[-1][0])):                       # eval padding at max dir level
        idcs_list.append(idcs_list[1])
    elif(pts is eval_pts and lvl_info[-1][0]==lvl_info[-1][1]):
        idcs_list.append(idcs_list[2])
    else:
        idcs_list.append(reduce_max_lvl(eval_pts,eval_idcs,s,eval_max_l,lvl_info[-1][0]))
    # idcs_list[0][1], idcs_list[2][1] = None, None   # TODO: we do not need these rev idcs, remove them!

    if(periodic):
        PBC_op, pbc_ws = gen_pbc_op(boxlens[0], theta, periodic_axes, p, pbc_lvls, pbc_no_monopole)
        if(pbc_lvls < 0): # optionally, only add non-ws boxes on level 0
            PBC_op = jnp.zeros(((p+1)**2,(p+1)**2))
        if(len(periodic_axes)==3 and not jnp.all(jnp.isclose(boxlens[0],boxlens[0][0,0]))):
            raise ValueError("3D PBC are currently only supported for cubic level zero boxes. You can manually set the dimensions of the level zero box by specifying L0_boxcen and L0_boxlen.")
    else:   # for open boundary conditions, we get no local expansion on level 0
        PBC_op = jnp.zeros(((p+1)**2,(p+1)**2))
        pbc_ws = []

    unqs, invs = precompute_combs(p)

    if mem_limit is None:
        mem_limit = 1 if pts.device.platform == 'cpu' else jnp.inf    # TODO determine this automatically - the L2 cache size works well for GPUs. Also, add a (small) mem_limit depending on the number of CPU threads.

    hierarchy = {"pts": pts,
                 "eval_pts": eval_pts,
                 "idcs": idcs_list,
                 "boxcenters": boxcenters,
                 "eval_boxcenters": eval_boxcenters,
                 "mpl_cnct": mpl_cnct,
                 "dir_cnct": dir_cnct,
                 "unqs": unqs, "invs": invs,
                 "lvl_info": lvl_info,
                 "img_cnct": img_cnct,
                 "s": s,
                 "p": p,
                 "theta": theta,
                 "periodic_axes": periodic_axes,
                 "pbc_lvls": pbc_lvls if periodic else 0,
                 "pbc_no_monopole": pbc_no_monopole,
                 "PBC_op": PBC_op,
                 "mem_limit": mem_limit
                }
    if(debug_info):
        hierarchy["boxlens"] = boxlens
        hierarchy["eval_boxlens"] = eval_boxlens
        hierarchy["pbc_ws"] = pbc_ws
    return hierarchy

@jax.jit
def handle_padding(pts, chrgs, eval_pts, idcs):
    r"""
    Generate the padded arrays required for computing the potential.
    """
    padded_pts = pts.at[idcs[0][0]].get(mode="fill",fill_value=0.0)   # TODO: we could buffer these arrays if desired...
    padded_chrgs = chrgs.at[idcs[0][0]].get(mode="fill",fill_value=0.0)  
    loc_padded_eval_pts = padded_pts if idcs[1] is idcs[0] else eval_pts.at[idcs[1][0]].get(mode="fill",fill_value=0.0)
    dir_padded_pts = padded_pts if idcs[2] is idcs[0] else pts.at[idcs[2][0]].get(mode="fill",fill_value=0.0)
    dir_padded_chrgs = padded_chrgs if idcs[2] is idcs[0] else chrgs.at[idcs[2][0]].get(mode="fill",fill_value=0.0)
    if(idcs[3] is idcs[2]):
        dir_padded_eval_pts = dir_padded_pts
    elif(idcs[3] is idcs[1]): 
        dir_padded_eval_pts = loc_padded_eval_pts
    else:
        dir_padded_eval_pts = eval_pts.at[idcs[3][0]].get(mode="fill",fill_value=0.0)
    return padded_pts, padded_chrgs, loc_padded_eval_pts, dir_padded_pts, dir_padded_chrgs, dir_padded_eval_pts


import jax.numpy as jnp
import jaxfmm.fmm as fmm
from jaxfmm.transf import *
import jaxfmm.debug_helpers as dbg
import pytest

@pytest.mark.parametrize('unitcube', [[128, {"N_max": 16}]], indirect=True)
def test_M2M(unitcube):
    chrgs, hier = unitcube
    pad_pts = hier["pts"][hier["idcs"][0][0],:]
    pad_chrgs = chrgs[hier["idcs"][0][0]]

    mpls = fmm.get_initial_mpls(pad_pts, pad_chrgs,hier["boxcenters"][1],hier["p"])
    mpls_merged = M2M(mpls,hier["boxcenters"][1],hier["boxcenters"][0],hier["unqs"],hier["invs"])
    mpls_direct = fmm.get_initial_mpls(pad_pts.reshape((1,-1,3)), pad_chrgs.reshape((1,-1)),hier["boxcenters"][0],hier["p"])
    assert jnp.allclose(mpls_merged, mpls_direct, rtol=1e-6, atol=1e-6)

@pytest.mark.parametrize('unitcube', [[128, {"N_max": 16}]], indirect=True)
def test_M2L(unitcube):
    chrgs, hier = unitcube
    to_shift = jnp.array([2.0,-4.0,5.0])

    mpls = fmm.get_initial_mpls(hier["pts"][None,:], chrgs[None,:],hier["boxcenters"][0],hier["p"])
    locs = M2L(hier["boxcenters"][0],hier["boxcenters"][0]+to_shift,jnp.array([[0]]),hier["unqs"],hier["invs"],mpls)
    locs_direct = dbg.get_locs(hier["pts"][None,:],chrgs[None,:],hier["boxcenters"][0]+to_shift,hier["p"])
    assert jnp.allclose(locs, locs_direct, rtol=1e-4, atol=1e-4)

@pytest.mark.parametrize('unitcube', [[128, {"N_max": 16}]], indirect=True)
def test_L2L(unitcube):
    chrgs, hier = unitcube
    pad_pts = hier["pts"][hier["idcs"][0][0],:]
    pad_chrgs = chrgs[hier["idcs"][0][0]]
    to_shift = jnp.array([2.0,-4.0,5.0])
    
    locs = dbg.get_locs(pad_pts.reshape((-1,3))+to_shift,pad_chrgs.flatten(),hier["boxcenters"][0],hier["p"])
    locs_dist = L2L(locs,hier["boxcenters"][0],hier["boxcenters"][1],hier["unqs"],hier["invs"])
    locs_direct = dbg.get_locs(pad_pts.reshape((1,-1,3))+to_shift, pad_chrgs.reshape((1,-1)),hier["boxcenters"][1],hier["p"])
    assert jnp.allclose(locs_dist, locs_direct,rtol=1e-6,atol=1e-2)


import jax
import jax.numpy as jnp
import jax.scipy as jsp
from jaxfmm.basis import inv_mpl_idx, eval_singular_basis
from time import perf_counter
import numpy as np
import pyvista as pv
import os
from functools import partial

__all__ = ["gen_stats", "gen_hierarchy_vtk", "gen_wellsep_vtk", "gen_pts_vtk", "time_function"]

def gen_stats(pts, eval_pts, idcs, boxcenters, eval_boxcenters, mpl_cnct, dir_cnct, img_cnct, lvl_info, s, periodic_axes, p, theta, print_stats = False, **kwargs):
    """Generate hierarchy stats. Use the hierarchy information from gen_hierarchy as input."""
    stats = {"p": p,
             "theta": theta,
             "max_l": len(boxcenters)-1,
             "eval_max_l": len(eval_boxcenters)-1,
             "num_pts": pts.shape[0],
             "num_eval_pts": eval_pts.shape[0],
             "num_child": 2**s,
             "pts_per_box": idcs[2][0].shape[1],
             "eval_pts_per_box": idcs[3][0].shape[1]}
    periodic = len(periodic_axes)>0
    mem_hierarch = idcs[0][0].nbytes + idcs[0][1].nbytes
    if(idcs[1] is not idcs[0]):
        mem_hierarch += idcs[1][0].nbytes + idcs[1][1].nbytes
    if(idcs[2] is not idcs[0]):
        mem_hierarch += idcs[2][0].nbytes + idcs[2][1].nbytes
    if((idcs[3] is not idcs[1]) and (idcs[3] is not idcs[2])):
        mem_hierarch += idcs[3][0].nbytes + idcs[3][1].nbytes
    mem_hierarch = dir_cnct.nbytes + img_cnct.nbytes
    M2L_nums = []
    M2L_pad_fracs = []
    n_img = img_cnct.shape[0]

    for i in range(len(lvl_info)-1):
        _, l_src = lvl_info[i]
        num_nopad = (mpl_cnct[i] < boxcenters[l_src].shape[0]*n_img).sum().tolist()
        M2L_nums.append(mpl_cnct[i].size)
        mem_hierarch += mpl_cnct[i].nbytes
        M2L_pad_fracs.append(1 if num_nopad==0 else M2L_nums[-1] / num_nopad)

    for i in range(stats["max_l"]):
        mem_hierarch += boxcenters[i].nbytes
        if 'boxlens' in kwargs:
            mem_hierarch += kwargs.get("boxlens")[i].nbytes

    if(eval_pts is not pts):
        for i in range(stats["eval_max_l"]):
            mem_hierarch += eval_boxcenters[i].nbytes
            if 'eval_boxlens' in kwargs:
                mem_hierarch += kwargs.get("eval_boxlens")[i].nbytes

    stats["M2L_nums"] = M2L_nums
    stats["lvl_info"] = lvl_info
    stats["M2L_pad_fracs"] = M2L_pad_fracs
    stats["direct_num"] = dir_cnct.size
    dir_no_pad = (dir_cnct < boxcenters[lvl_info[-1][1]].shape[0]*n_img).sum().tolist()
    stats["direct_pad_frac"] = (dir_cnct.size / dir_no_pad) if dir_no_pad != 0 else 1
    stats["compression_ratio"] = dir_no_pad * idcs[3][0].shape[1] * idcs[2][0].shape[1] / (pts.shape[0]*eval_pts.shape[0])
    stats["mem_hierarch"] = mem_hierarch
    stats["mem_coeffs"] = 4*2*((p+1)**2)*(8**(stats["max_l"]+1)-1)/7 # 4 bytes, 2 arrays, (p+1)**2 coeffs and nboxes
    stats["mem_pts_chrgs"] = (pts.nbytes*4)/3
    if(eval_pts is not pts):
        stats["mem_pts_chrgs"] += eval_pts.nbytes
    if(print_stats):
        print("---------------------FMM Hierarchy Stats---------------------")
        print("p = %i, theta = %.2f, %i children per box"%(p,theta,stats["num_child"]))
        if(periodic): 
            print("Periodic boundary on axes: ", periodic_axes)
        print("%i points, %i levels, %i particles per box"%(stats["num_pts"],stats["max_l"], stats["pts_per_box"]))
        if(eval_pts is not pts):
            print("%i eval points, %i eval levels, %i eval points per box"%(stats["num_eval_pts"],stats["eval_max_l"],stats["eval_pts_per_box"]))
        print("")
        print(" (l_eval, l_src) | padding frac. | interactions")
        print("-----------------------------------------------")
        for i in range(len(lvl_info)-1):
            l_eval, l_src = lvl_info[i]
            print(" M2L (%2i,%2i)     |     %5.2f     | %11i"%(l_eval, l_src, stats["M2L_pad_fracs"][i], stats["M2L_nums"][i]))
        totsum = jnp.array(stats["M2L_nums"]).sum()
        print("-----------------------------------------------")
        print(" M2L Total:            %5.2f       %11i"%(totsum/jnp.array(stats["M2L_nums"]).dot(1/jnp.array(stats["M2L_pad_fracs"])),totsum))
        print("-----------------------------------------------")
        print(" dir (%2i,%2i)     |     %5.2f     | %11i"%(lvl_info[-1][0], lvl_info[-1][1], stats["direct_pad_frac"], stats["direct_num"]))
        print("-----------------------------------------------")
        print("\nDirect interaction compression (without padding): %.2e"%(stats["compression_ratio"]))
        print("\nMemory used by the hierarchy:            %.2e Bytes"%stats["mem_hierarch"])
        print("Memory used by mpl + local coeffs (p=%i): %.2e Bytes"%(p, stats["mem_coeffs"])) 
        print("Memory used by the points + charges:     %.2e Bytes"%stats["mem_pts_chrgs"])
        print("-------------------------------------------------------------")
    return stats

def make_pyvista_mesh(boxcenters, boxlens):
    r"""
    Generate pyvista mesh of boxes with centers at boxcenters and diagonals boxlens.
    """
    pv_shifts = np.array([[-1,-1,-1],[1,-1,-1],[1,1,-1],[-1,1,-1],[-1,-1,1],[1,-1,1],[1,1,1],[-1,1,1]])
    pts = np.zeros((boxcenters.shape[0],8,3))
    for i in range(pv_shifts.shape[0]):
        pts[:,i,:] = boxcenters + boxlens/2 * pv_shifts[i,None,:]
    pts = pts.reshape((-1,3))
    cells = np.arange(pts.shape[0],dtype=np.int32).reshape((-1,8))
    return pv.UnstructuredGrid({pv.CellType.HEXAHEDRON: cells}, pts)

def gen_hierarchy_vtk(boxcenters, boxlens, eval_boxcenters, eval_boxlens, dir="hierarchy", **kwargs):
    r"""
    Output a series of vtk files showing the FMM hierarchy on every level. Does not show virtual PBC images.
    """
    eval_max_l = len(eval_boxcenters)
    max_l = len(boxcenters)
    if not os.path.exists(dir):
        os.makedirs(dir)
    for l in range(max_l):
        mesh = make_pyvista_mesh(boxcenters[l],boxlens[l])
        mesh.cell_data["src_level"] = np.ones(mesh.points.shape[0]//8)*l
        mesh.save("%s/src_level_%i.vtk"%(dir,l))
    
    if(eval_boxcenters is not boxcenters or eval_boxlens is not boxlens):
        for l in range(eval_max_l):
            mesh = make_pyvista_mesh(eval_boxcenters[l],eval_boxlens[l])
            mesh.cell_data["eval_level"] = np.ones(mesh.points.shape[0]//8)*l
            mesh.save("%s/eval_level_%i.vtk"%(dir,l))

def gen_pts_vtk(pts, eval_pts, dir="hierarchy", **kwargs):
    r"""
    Output the hierarchy points as vtk.
    """
    if not os.path.exists(dir):
        os.makedirs(dir)
    point_cloud = pv.PolyData(np.array(pts))
    point_cloud.save("%s/src_pts.vtk"%(dir))
    point_cloud = pv.PolyData(np.array(eval_pts))
    point_cloud.save("%s/eval_pts.vtk"%(dir))

def gen_wellsep_vtk(id, boxcenters, boxlens, eval_boxcenters, eval_boxlens, mpl_cnct, dir_cnct, img_cnct, lvl_info, s, pbc_ws, dir="hierarchy", **kwargs):
    r"""
    Output a vtk file showing all the boxes that are considered for potential calculation, given an id of a box on the highest level. Includes virtual PBC images.
    """
    if not os.path.exists(dir):
        os.makedirs(dir)
    max_l_eval = lvl_info[-1][0]
    n_chi = 2**s
    periodic = len(pbc_ws)>0
    meshlist = []

    L = boxlens[0][0]
    for l, img_cnct_ws in enumerate(pbc_ws):
        mesh = make_pyvista_mesh(boxcenters[0] - img_cnct_ws[1],jnp.tile(L,(img_cnct_ws[1].shape[0],1)))
        mesh.cell_data["eval_level"] = np.ones(mesh.points.shape[0]//8)*(-l)
        mesh.cell_data["src_level"] = np.ones(mesh.points.shape[0]//8)*(-l)
        meshlist.append(mesh)
        L += jnp.ptp(img_cnct_ws[0],axis=0)

    for i in range(len(lvl_info)-1):
        l_eval, l_src = lvl_info[i]
        extract = mpl_cnct[i][id//(n_chi**(max_l_eval-l_eval))]
        n_src = n_chi**l_src
        if(periodic):
            extract = extract[extract < (n_src*img_cnct.shape[0])]
            mesh = make_pyvista_mesh(boxcenters[l_src][extract%n_src] + img_cnct[extract//n_src],boxlens[l_src][extract%n_src])
        else:
            extract = extract[extract < n_src]
            mesh = make_pyvista_mesh(boxcenters[l_src][extract],boxlens[l_src][extract])
        mesh.cell_data["eval_level"] = np.ones(mesh.points.shape[0]//8)*(l_eval)
        mesh.cell_data["src_level"] = np.ones(mesh.points.shape[0]//8)*(l_src)
        meshlist.append(mesh)

    l_eval, l_src = lvl_info[-1]
    extract = dir_cnct[id]
    n_src = n_chi**l_src
    if(periodic):
        extract = extract[extract<n_src*img_cnct.shape[0]]
        mesh = make_pyvista_mesh(boxcenters[-1][extract%n_src] + img_cnct[extract//n_src],boxlens[-1][extract%n_src])
    else:
        extract = extract[extract < n_src]
        mesh = make_pyvista_mesh(boxcenters[l_src][extract],boxlens[l_src][extract])

    mesh.cell_data["eval_level"] = np.ones(mesh.points.shape[0]//8)*(l_eval + 1)
    mesh.cell_data["src_level"] = np.ones(mesh.points.shape[0]//8)*(l_src + 1)
    meshlist.append(mesh)

    mesh = make_pyvista_mesh(eval_boxcenters[l_eval][id][None,:],eval_boxlens[l_eval][id][None,:]*1.001)   # the multiplication fixes pyvista getting confused for eval_pts = pts
    mesh.cell_data["eval_level"] = np.ones(1)*(l_eval + 2)
    mesh.cell_data["src_level"] = np.ones(1)*(l_src + 2)
    meshlist.append(mesh)

    mergedmesh = pv.merge(meshlist)
    mergedmesh.save("%s/wellsep_%i.vtk"%(dir,id))

@partial(jax.jit, static_argnames=['p'])
def eval_mpls(mpls, eval_pts, boxcenters, p):
    r"""
    Evaluate multipole expansions.
    """
    sing = eval_singular_basis(eval_pts - boxcenters[None,:],p)
    ms, _ = inv_mpl_idx(jnp.arange((p+1)**2))
    prefac = ((2-(ms==0))*(1-2*(ms<0)))[None,None,:]
    res = (mpls * sing * prefac).sum(axis=2)
    return res / (4*jnp.pi)

@partial(jax.jit, static_argnames=['p'])
def get_locs(padded_pts, padded_chrgs, boxcenters, p):
    r"""
    Obtain local expansions directly.
    """
    dist = padded_pts - boxcenters[:,None]
    sing = eval_singular_basis(dist,p)
    return (sing * padded_chrgs[...,None]).sum(axis=1)

def binom(x, y):
  return jnp.exp(jsp.special.gammaln(x + 1) - jsp.special.gammaln(y + 1) - jsp.special.gammaln(x - y + 1))

def gen_multipole_dist(m, n, eps = 0.5):
    r"""
    Generate a point charge distribution corresponding to a specific multipole moment (Majic, Matt. (2022). 
    Point charge representations of multipoles. European Journal of Physics. 43. 10.1088/1361-6404/ac578b.)
    """
    if(m == 0):   # axial
        k = jnp.arange(-n, n+1, 2)
        chrgs = (-1)**((n-k)/2) * binom(n, (n-k)/2.0) / (jsp.special.factorial(n) * (2*eps)**n)
        pts = jnp.zeros((k.shape[0],3))
        pts = pts.at[:,2].set(k*eps)
    else:         # (stacked) bracelet
        rotate = m < 0
        m = abs(m)      # we work with the real basis and rotate later
        knum = n-m+1
        jnum = 2*m
        j = jnp.tile(jnp.arange(jnum),knum)
        k = jnp.repeat(jnp.arange(-n+m,n-m+1,2),jnum)
        phi = (j-0.5) * jnp.pi/m if rotate else j * jnp.pi/m
        pts = jnp.array([eps*jnp.cos(phi), eps*jnp.sin(phi), k*eps]).T
        chrgs = 4**(m-1) * jsp.special.factorial(m-1) / ((2*eps)**n * jsp.special.factorial(n-m)) * (-1)**((n-m-k)/2 + j) * binom(n-m,(n-m-k)/2)
    return pts, chrgs

def time_function(func, nruns = 10, print_times = True, **kwargs):
    r"""
    Time function func, obtaining the best of nruns runs and optionally printing the results. Function args are supplied as additional kwargs.
    """
    best = jnp.inf
    for i in range(nruns):
        T0 = perf_counter()
        res = jax.block_until_ready(func(**kwargs))
        T1 = perf_counter()
        best = min(T1-T0,best)
        if(i==0):
            comp = best
    if(print_times):
        print("%s compilation: %.2e s"%(func.__name__,comp))
        print("%s runtime: %.2e s"%(func.__name__,best))
    return res, comp, best


import jax.numpy as jnp
from jaxfmm.basis import mpl_idx, inv_mpl_idx
import jax
from jaxfmm.basis import eval_regular_basis, eval_singular_basis
from functools import partial

def gen_img_connectivity(L0_boxlen, theta, periodic_axes):
    r"""
    Similar to gen_connectivity, but computes connectivity information for periodic images.
    The generated patterns are rectangular cuboids, which helps mitigate numerical inaccuracies for the resulting PBC operator.
    """
    R = jnp.linalg.norm(L0_boxlen/2)                           # r = R in this case, all boxes are the same
    num_images = jnp.int32((1+theta)/(theta*L0_boxlen) * R)[0] # lower bound of non-ws images in axis directions
    non_per = jnp.array([False if i in periodic_axes else True for i in range(3)])
    num_images = num_images.at[non_per].set(0)
    mul_facs = 2*num_images + 1                                # total number of boxes in axis directions
    ids = [slice(-num_images[i],num_images[i]+1) if i in periodic_axes else slice(0,1) for i in range(3)]
    img_ids = jnp.reshape(jnp.mgrid[ids].T,(-1,3))  # spawn all the images

    L0_boxlen_up = L0_boxlen * mul_facs   # boxlength on the parent level
    R_up = jnp.linalg.norm(L0_boxlen_up/2)
    num_images_up = jnp.int32((1+theta)/(theta*L0_boxlen_up) * R_up)[0] # upper bound of ws images in axis directions
    ids_up = [slice(-num_images_up[i],num_images_up[i]+1) if i in periodic_axes else slice(0,1) for i in range(3)]
    img_ids_up = jnp.reshape(jnp.mgrid[ids_up].T,(-1,3))                                # spawn parent images
    img_ids_up = img_ids_up[jnp.any(img_ids_up!=0,axis=-1)]                             # we take out the non-ws area
    img_ids_up = ((mul_facs*img_ids_up)[:,None,:] + img_ids[None,:,:]).reshape((-1,3))  # spawn children for each remaining parent

    return img_ids*L0_boxlen, img_ids_up*L0_boxlen, mul_facs

@partial(jax.jit, static_argnames = ["p"])
def gen_M2Mop(p, dists, reg_zeros = []):    # TODO: generate via rotations
    r"""
    Generate M2M operator matrix for PBC.
    """
    reg = eval_regular_basis(dists,p).sum(axis=0)   # sign of dists does not matter here (symmetry)
    reg = reg.at[reg_zeros].set(0)  # manually set terms to zero that should be zero (symmetry)
    M2Mop = jnp.zeros(((p+1)**2,(p+1)**2))

    for j in range(p+1):
        for k in range(0,j+1):  # Real coeffs
            for n in range(j+1):
                for m in range(max(k+n-j,-n),min(k+j-n,n)+1):
                    M2Mop = M2Mop.at[mpl_idx(k,j),mpl_idx(abs(k-m),j-n)].add((-1)**((abs(k)-abs(m)-abs(k-m))//2) * reg[...,mpl_idx(abs(m),n)])
                    M2Mop = M2Mop.at[mpl_idx(k,j),mpl_idx(-abs(k-m),j-n)].add(-(-1)**((abs(k)-abs(m)-abs(k-m))//2) * jnp.sign(m)*jnp.sign(k-m)*reg[mpl_idx(-abs(m),n)])
        for k in range(-j,0):   # Imag coeffs
            for n in range(j+1):
                for m in range(max(k+n-j,-n),min(k+j-n,n)+1):
                    M2Mop = M2Mop.at[mpl_idx(k,j),mpl_idx(-abs(k-m),j-n)].add(-(-1)**((abs(k)-abs(m)-abs(k-m))//2) * jnp.sign(k-m)*reg[...,mpl_idx(abs(m),n)])
                    M2Mop = M2Mop.at[mpl_idx(k,j),mpl_idx(abs(k-m),j-n)].add(-(-1)**((abs(k)-abs(m)-abs(k-m))//2) * jnp.sign(m)*reg[...,mpl_idx(-abs(m),n)])

    return M2Mop

@partial(jax.jit, static_argnames = ["p"])
def gen_M2Lop(p, dists, sing_zeros = []): # TODO: generate via rotations
    r"""
    Generate M2L operator matrix for PBC.
    """
    sing = eval_singular_basis(dists,2*p).sum(axis=0)   # sign of dists does not matter here (symmetry)
    sing = sing.at[sing_zeros].set(0) # manually set terms to zero that should be zero (symmetry)
    M2Lop = jnp.zeros(((p+1)**2,(p+1)**2))

    for j in range(p+1):
        for k in range(1,j+1):  # -Imag coeffs!
            for n in range(p+1):
                for m in range(max(k-j-n,-n),min(j+n+k,n)+1):
                    M2Lop = M2Lop.at[mpl_idx(k,j),mpl_idx(abs(m),n)].add((-1)**((abs(k-m)-abs(k)-abs(m))//2) * (-jnp.sign(m-k)*sing[mpl_idx(-abs(m-k),j+n)]))
                    M2Lop = M2Lop.at[mpl_idx(k,j),mpl_idx(-abs(m),n)].add((-1)**((abs(k-m)-abs(k)-abs(m))//2) * jnp.sign(m)*sing[mpl_idx(abs(m-k),j+n)])
        for k in range(-j,1):   # Real coeffs!
            for n in range(p+1):
                for m in range(max(k-j-n,-n),min(j+n+k,n)+1):
                    M2Lop = M2Lop.at[mpl_idx(k,j),mpl_idx(abs(m),n)].add((-1)**((abs(k-m)-abs(k)-abs(m))//2) * sing[mpl_idx(abs(m-k),j+n)])
                    M2Lop = M2Lop.at[mpl_idx(k,j),mpl_idx(-abs(m),n)].add((-1)**((abs(k-m)-abs(k)-abs(m))//2) * jnp.sign(m)* jnp.sign(m-k) * sing[mpl_idx(-abs(m-k),j+n)])
    return M2Lop

def gen_pbc_op(L, theta, periodic_axes, p, pbc_lvls, pbc_no_monopole):
    r"""
    Generate full PBC operator matrix.
    """
    par_zero = [jnp.zeros(((i+1)*p+1)**2,dtype=jnp.uint8) for i in range(2)]
    for i, func in enumerate((eval_regular_basis,eval_singular_basis)):
        par = func(jnp.array([1,2,3]),(i+1)*p)
        uneven = jnp.any(jnp.stack((func(jnp.array([-1,2,3]),(i+1)*p)/par < 0,
                                    func(jnp.array([1,-2,3]),(i+1)*p)/par < 0,
                                    func(jnp.array([1,2,-3]),(i+1)*p)/par < 0),axis=-1),axis=-1)  # as long as the pattern is symmetric, these cancel
        #cubic = jnp.abs(func(jnp.array([1,1,1]),p))<1e-6   # these components only cancel for patterns that are both symmetric and cubic
        par_zero[i] = jnp.where(par_zero[i].at[uneven].set(1))[0]
    reg_zeros, sing_zeros = par_zero

    img_non_ws, img_ws, mul_facs = gen_img_connectivity(L,theta,periodic_axes)
    M2Mop = jnp.eye((p+1)**2)
    res = gen_M2Lop(p,img_ws,sing_zeros)
    img_cnct_list = [[img_non_ws, img_ws]]
    for i in range(1,pbc_lvls+1):
        L *= mul_facs
        img_non_ws, img_ws, mul_facs = gen_img_connectivity(L,theta,periodic_axes)
        M2Mop = gen_M2Mop(p, img_non_ws,reg_zeros)@M2Mop
        M2Lop = gen_M2Lop(p, img_ws,sing_zeros)
        res += M2Lop@M2Mop
        img_cnct_list.append([img_non_ws, img_ws])
    if(pbc_no_monopole):    # set monopole contribution to zero
        res = res.at[0,0].set(0)

    ### edit the pbc op such that it works with the new basis TODO: deal with this in M2Lop, M2Mop
    ms, ns = inv_mpl_idx(jnp.arange((p+1)**2))
    flipidx = mpl_idx(-ms,ns)
    res *= (-1+2*(ms<0))**ns[:,None]
    res = res[flipidx,:]

    return res, img_cnct_list


import jax
import jax.numpy as jnp
from functools import partial
from jax.scipy.special import factorial as fac
from jaxfmm.basis import mpl_idx, inv_mpl_idx

### TODO: make rotations faster (their cost still dominates the transformation operators)

@jax.jit
def cart_to_sph(rvec):
    r"""
    Convert cartesian to spherical coordinates.
    """
    xy = rvec[...,0]**2 + rvec[...,1]**2
    r = jnp.sqrt(xy + rvec[...,2]**2)
    theta = jnp.arctan2(jnp.sqrt(xy),rvec[...,2])
    phi = jnp.arctan2(rvec[...,1],rvec[...,0])
    return jnp.stack((r,theta,phi),axis=-1)
    
def precompute_combs(p):
    r"""
    Precompute unique Jacobi polynomial parameters that appear in Wigner rotation matrices up to expansion order p.
    Returns the unique combinations and their positions in the rotation matrices.
    """
    unqs, invs, unq_num = [], [], [0]

    for j in range(p+1):
        rng = jnp.arange(-j,j+1)    # m and k
        mu = jnp.abs(rng[:,None]-rng[None,:]).flatten()
        nu = jnp.abs(rng[:,None]+rng[None,:]).flatten()
        s = j - (mu+nu)//2
        comb = jnp.stack((s,mu,nu),axis=0)
        unq, inv = jnp.unique(comb,axis=1,return_inverse=True)
        unqs.append(unq)
        invs.append(inv)
        unq_num.append(unq_num[j] + unq.shape[1])

    ### combine the unqs into one array
    unqs_tot = jnp.concatenate(unqs,axis=-1)
    resrt = jnp.argsort(unqs_tot[0,...])
    unqs_tot = unqs_tot[:,resrt]

    ### reindex the invs
    inv_rsrt = jnp.argsort(resrt)
    for j in range(p+1):
        invs[j] = inv_rsrt[invs[j]+unq_num[j]].reshape((2*j+1,2*j+1))[j:,:]

    return unqs_tot, invs

def cut(i, p):   # (p-i)^2
    r"""
    Number of Jacobi polynomial parameter combinations with s==i for expansion order p.
    """
    return (p+1-i)**2

def cut_cum(i, p):   # sum(n^2)_n=(p-i)^p
    r"""
    Number of Jacobi polynomial parameter combinations with s<=i for expansion order p.
    """
    return ((p+1)*(p+2)*(2*p+3) - (p-i)*(p+1-i)*(2*(p-i)+1))//6

@partial(jax.jit, static_argnames=['j'])
def fast_jacobi(j, mu, nu, x):
    r"""
    Compute Jacobi polynomials P^(mu,nu)_s(x) with a recursion relation 
    (adapted from https://doi.org/10.1103/PhysRevC.106.054320).
    """
    res = jnp.ones((*x.shape,mu.shape[0])) # res[...,:cut_cum(0,j)] = 1
    mu = mu[None,None,...]
    nu = nu[None,None,...]
    x = x[...,None]

    if(j>0):
        mu = mu[...,cut(0,j):]
        nu = nu[...,cut(0,j):]
        res1 = res[...,cut_cum(0,j):]

        res = res.at[...,cut_cum(0,j):].set(0.5*(2*(mu+1)+(mu+nu+2)*(x-1)))
    
    for i in range(2,j+1):
        mu = mu[...,cut(i-1,j):]
        nu = nu[...,cut(i-1,j):]
        res2 = res1[...,cut(i-1,j):]
        res1 = res[...,cut_cum(i-1,j):]

        xn = i - 1
        two_nab = 2*xn + mu + nu
        a1 = 2*(xn+1) * (xn+mu+nu+1) * two_nab
        a2 = (two_nab+1) * (mu**2 - nu**2)
        a3 = two_nab * (two_nab+1) * (two_nab+2)
        a4 = 2*(xn+mu) * (xn+nu) * (two_nab+2)
        res = res.at[...,cut_cum(i-1,j):].set(((a2+a3*x)*res1 - a4*res2)/a1)
    return res

def factorialfac(s, mu, nu):
    return jnp.sqrt(fac(s)*fac(s+mu+nu) / (fac(s+mu)*fac(s+nu)))

@partial(jax.jit, static_argnames=['p'])
def get_polar_rot_coeff(p, unqs, theta):
    r"""
    Compute all unique Wigner rotation matrix coefficients up until expansion order p.
    """
    basemat = fast_jacobi(p, unqs[1], unqs[2], jnp.cos(theta))
    basemat *= factorialfac(unqs[0], unqs[1], unqs[2])[None,None,...] * \
               jnp.sin(theta/2)[...,None]**unqs[1][None,None,...] * \
               jnp.cos(theta/2)[...,None]**unqs[2][None,None,...]
    return basemat

def Nfac(k, l, m):
    return jnp.sqrt(fac(l-k)*fac(l+k) / (fac(l-m)*fac(l+m)))

def eps(m, k):
    return (-1)**((m-k) * (k < m))

@jax.jit
def rot_polar(basemat, invs, coeffs, inv=False, loc=False):
    r"""
    Rotate multipole/local coefficients around the y-axis.
    """
    p = invs[-1].shape[0] - 1
    k = jnp.arange(-p,p+1)
    for j in range(p+1):
        ### assemble rotation matrix
        rotmats = basemat[...,invs[j]]
        fact = (eps(k[None,p-j:p+j+1], k[p:p+j+1,None]) *                # missing factor from the recursion
                Nfac(k[p:p+j+1,None],j,k[None,p-j:p+j+1])**(2*loc-1) *   # normalization of the real basis
                (2*inv-1)**(k[None,p-j:p+j+1] + k[p:p+j+1,None]))        # inverse rotation (theta -> -theta)
        rotmats *= fact[None,None,...]
    
        ### real update part
        real_update = rotmats[...,j:]
        imag_update = rotmats[...,:j]*((-1)**jnp.arange(j,2*j))[None,:] # invert signs for m<0
        real_update = real_update.at[...,1:].add(jnp.flip(imag_update,axis=-1))
        coeffs = coeffs.at[...,mpl_idx(0,j):mpl_idx(j,j)+1].set(jnp.einsum("ijmk, ijk -> ijm",real_update,coeffs[...,mpl_idx(0,j):mpl_idx(j,j)+1]))
        
        ### imag update part
        imag_update = jnp.flip(rotmats[...,1:,j+1:],axis=(-2,-1)) - jnp.flip(imag_update[...,1:,:],axis=(-2))    # flipped to match the real basis ordering
        coeffs = coeffs.at[...,mpl_idx(-j,j):mpl_idx(0,j)].set(jnp.einsum("ijmk, ijk -> ijm",imag_update,coeffs[...,mpl_idx(-j,j):mpl_idx(0,j)]))
    return coeffs

@jax.jit
def rot_azimuth(coeffs, phi, inv=False, loc=False):
    r"""
    Rotate multipole/local coefficients around the z-axis.
    """
    ms, ns = inv_mpl_idx(jnp.arange(coeffs.shape[-1]))
    flipidx = mpl_idx(-ms,ns)
    sgns = (1-2*inv)*(1-2*loc)
    mphi = ms[None,None,:]*phi[...,None]

    coeffs = coeffs * jnp.cos(mphi) - sgns*coeffs[...,flipidx] * jnp.sin(mphi)   # NOTE: we intentionally swap the sign for sin
    return coeffs


import jax
import jax.numpy as jnp
from functools import partial

def mpl_idx(m, n):
    r"""
    Compute flattened array position of multipole coefficient C^m_n with order m and degree n.
    """
    return n**2 + (m+n)

def inv_mpl_idx(idx):
    r"""
    Compute order m and degree n of flattened array position idx.
    """
    n = jnp.int32(jnp.floor(jnp.sqrt(idx)+1e-4))
    m = idx - n*(n+1)
    return m, n

@partial(jax.jit, static_argnames=['p'])
def eval_regular_basis(rvec, p):    # TODO: swap signs again and check if this removes the (-1)^m in the rotation matrix?
    r"""
    Evaluate real regular basis functions (Laplace kernel) with a recursion relation, based on 
    [Gumerov, N. A. et al. Fast multipole methods on graphics processors. J. Comp. Phys., B 227, 8290 (2008)].
    """
    x, y, z = rvec[...,0], rvec[...,1], rvec[...,2]
    r2 = (rvec**2).sum(axis=-1)
    coeff = jnp.ones((*rvec.shape[:-1], (p+1)**2))  # coeff[...,mpl_idx(0,0)] = 1

    if(p>0):
        coeff = coeff.at[...,mpl_idx(-1,1)].set(-0.5*y)
        coeff = coeff.at[...,mpl_idx(0,1)].set(z)
        coeff = coeff.at[...,mpl_idx(1,1)].set(0.5*x)
    for n in range(2,p+1):    # first/second recursion: extreme values and their neighbors
        coeff = coeff.at[...,mpl_idx(-n,n)].set((x*coeff[...,mpl_idx(1-n,n-1)] - y*coeff[...,mpl_idx(n-1,n-1)])/(2*n))
        coeff = coeff.at[...,mpl_idx(1-n,n)].set(z*coeff[...,mpl_idx(1-n,n-1)])
        for m in range(2-n,n-1):   # third recursion: all values inbetween
            coeff = coeff.at[...,mpl_idx(m,n)].set(((2*n-1)*z*coeff[...,mpl_idx(m,n-1)] - 
                                                    r2*coeff[...,mpl_idx(m,n-2)])/((n-abs(m))*(n+abs(m))))
        coeff = coeff.at[...,mpl_idx(n-1,n)].set(z*coeff[...,mpl_idx(n-1,n-1)])
        coeff = coeff.at[...,mpl_idx(n,n)].set((x*coeff[...,mpl_idx(n-1,n-1)] + y*coeff[...,mpl_idx(1-n,n-1)])/(2*n))
    return coeff

@partial(jax.jit, static_argnames=['p'])
def eval_regular_basis_grad(rvec, p):
    r"""
    Evaluate the gradient of real regular basis functions (Laplace kernel) with a recursion relation.
    """
    scal_coeff = eval_regular_basis(rvec, p)
    x, y, z = rvec[...,0], rvec[...,1], rvec[...,2]
    coeff = jnp.zeros((*rvec.shape[:-1], (p+1)**2, 3)) # coeff[...,mpl_idx(0,0),:] = 0

    if(p>0):
        coeff = coeff.at[...,mpl_idx(1,1),0].set(0.5)
        coeff = coeff.at[...,mpl_idx(0,1),2].set(1.0)
        coeff = coeff.at[...,mpl_idx(-1,1),1].set(-0.5)
    for n in range(2,p+1):    
        # first recursion: extreme values
        coeff = coeff.at[...,mpl_idx(n,n),:].set(x[...,None]*coeff[...,mpl_idx(n-1,n-1),:] + 
                                                 y[...,None]*coeff[...,mpl_idx(1-n,n-1),:])
        coeff = coeff.at[...,mpl_idx(n,n),0].add(scal_coeff[...,mpl_idx(n-1,n-1)])
        coeff = coeff.at[...,mpl_idx(n,n),1].add(scal_coeff[...,mpl_idx(1-n,n-1)])
        coeff = coeff.at[...,mpl_idx(n,n),:].divide(2*n)

        coeff = coeff.at[...,mpl_idx(-n,n),:].set(x[...,None]*coeff[...,mpl_idx(1-n,n-1),:] - 
                                                  y[...,None]*coeff[...,mpl_idx(n-1,n-1),:])
        coeff = coeff.at[...,mpl_idx(-n,n),0].add(scal_coeff[...,mpl_idx(1-n,n-1)])
        coeff = coeff.at[...,mpl_idx(-n,n),1].subtract(scal_coeff[...,mpl_idx(n-1,n-1)])
        coeff = coeff.at[...,mpl_idx(-n,n),:].divide(2*n)

        # second recursion: neighbors of extreme values
        coeff = coeff.at[...,mpl_idx(n-1,n),:].set(z[...,None]*coeff[...,mpl_idx(n-1,n-1),:])
        coeff = coeff.at[...,mpl_idx(n-1,n),2].add(scal_coeff[...,mpl_idx(n-1,n-1)])
        
        coeff = coeff.at[...,mpl_idx(1-n,n),:].set(z[...,None]*coeff[...,mpl_idx(1-n,n-1),:])
        coeff = coeff.at[...,mpl_idx(1-n,n),2].add(scal_coeff[...,mpl_idx(1-n,n-1)])

        # third recursion: all values inbetween
        for m in range(-n+2,n-1):
            coeff = coeff.at[...,mpl_idx(m,n),:].set((2*n-1)*z[...,None]*coeff[...,mpl_idx(m,n-1),:] - 
                                                     (rvec**2).sum(axis=-1)[...,None]*coeff[...,mpl_idx(m,n-2),:] - 
                                                     2*scal_coeff[...,mpl_idx(m,n-2),None]*rvec)
            coeff = coeff.at[...,mpl_idx(m,n),2].add((2*n-1)*scal_coeff[...,mpl_idx(m,n-1)])
            coeff = coeff.at[...,mpl_idx(m,n),:].divide((n-abs(m)) * (n+abs(m)))
    return coeff

@partial(jax.jit, static_argnames=['p'])
def eval_singular_basis(rvec, p):
    r"""
    Evaluate real singular basis functions (Laplace kernel) with a recursion relation.
    """
    x, y, z = rvec[...,0], rvec[...,1], rvec[...,2]
    r2 = (rvec**2).sum(axis=-1)
    coeff = jnp.zeros((*rvec.shape[:-1], (p+1)**2))
    coeff = coeff.at[...,mpl_idx(0,0)].set(1/jnp.sqrt(r2))

    if(p>0):
        coeff = coeff.at[...,mpl_idx(1,1)].set(coeff[...,mpl_idx(0,0)]*x/r2)
        coeff = coeff.at[...,mpl_idx(0,1)].set(coeff[...,mpl_idx(0,0)]*z/r2)
        coeff = coeff.at[...,mpl_idx(-1,1)].set(coeff[...,mpl_idx(0,0)]*y/r2)
    for n in range(2,p+1): # first/second recursion: extreme values and their neighbors
        coeff = coeff.at[...,mpl_idx(n,n)].set((2*n-1)*(x*coeff[...,mpl_idx(n-1,n-1)] - 
                                                        y*coeff[...,mpl_idx(1-n,n-1)])/r2)
        coeff = coeff.at[...,mpl_idx(n-1,n)].set((2*n-1)*z*coeff[...,mpl_idx(n-1,n-1)]/r2)
        for m in range(2-n,n-1): # third recursion: all values inbetween
            coeff = coeff.at[...,mpl_idx(m,n)].set(((2*n-1)*z*coeff[...,mpl_idx(m,n-1)] - 
                                                    (n-1-m)*(n-1+m)*coeff[...,mpl_idx(m,n-2)])/r2)
        coeff = coeff.at[...,mpl_idx(1-n,n)].set((2*n-1)*z*coeff[...,mpl_idx(1-n,n-1)]/r2)
        coeff = coeff.at[...,mpl_idx(-n,n)].set((2*n-1)*(x*coeff[...,mpl_idx(1-n,n-1)] + 
                                                         y*coeff[...,mpl_idx(n-1,n-1)])/r2)
    return coeff


import jax.numpy as jnp
from jax import random
from jaxfmm import *

print("Generating sample points and charges.")
N = 2**15
key = random.key(124)
pts = random.uniform(key,(N,3))
chrgs = random.uniform(key,N)

print("Generating FMM hierarchy.")
tree_info = gen_hierarchy(pts)

print("Compiling and computing FMM potential.")
pot_FMM = eval_potential(chrgs, **tree_info)

print("Compiling and computing analytic potential.")
pot_dir = eval_potential_direct(pts, chrgs)

print("FMM normwise relative error: %.2e"%(jnp.linalg.norm(pot_dir-pot_FMM)/jnp.linalg.norm(pot_dir)))

### if dev dependencies (jaxfmm[dev]) are installed, debug information can be printed as follows:
# from jaxfmm.debug_helpers import *
# tree_info = gen_hierarchy(pts, debug_info = True)
# gen_stats(**tree_info, print_stats = True)

### additionally, vtk visualizations can be generated as follows:
# gen_hierarchy_vtk(**tree_info)
# gen_wellsep_vtk(0, **tree_info)


import pytest
from jaxfmm.rotation import *
import jaxfmm.fmm as fmm
import jax.numpy as jnp
from jax.scipy.spatial.transform import Rotation
import jaxfmm.debug_helpers as dbg

@pytest.mark.parametrize('unitcube, phi, theta', [((128, {}), 0.6, 1.9), ((128, {}), 4.4, 5.7)], indirect=["unitcube"])
def test_mpl_rot(unitcube, phi, theta): # TODO: remove the hierarchy, we do not really need it here...
    chrgs, hier = unitcube
    p, unqs, invs = hier["p"], hier["unqs"], hier["invs"]

    # polar matrix coeffs
    basemat = get_polar_rot_coeff(p,unqs,jnp.array([[theta]]))

    # initial mpls
    mpls_og = fmm.get_initial_mpls(hier["pts"][None,...], chrgs[None,...],hier["boxcenters"][0],hier["p"])

    mpls = rot_azimuth(mpls_og,jnp.array([[phi]]))
    mpls = rot_polar(basemat,invs,mpls)
    mpls_back = rot_polar(basemat,invs,mpls,True)
    mpls_back = rot_azimuth(mpls_back,jnp.array([[phi]]),True)

    spacerot = Rotation.from_euler('zy', jnp.array([-phi,-theta]))
    mpls_manualrot = fmm.get_initial_mpls(spacerot.apply(hier["pts"])[None,...], chrgs[None,...],hier["boxcenters"][0],p)

    assert jnp.allclose(mpls_og,mpls_back,rtol=1e-2,atol=1e-2)      # test inverse rotation
    assert jnp.allclose(mpls_manualrot,mpls,rtol=1e-2,atol=1e-2)    # test forward rotation

@pytest.mark.parametrize('unitcube, phi, theta', [((128, {}), 0.6, 1.9), ((128, {}), 4.4, 5.7)], indirect=["unitcube"])
def test_loc_rot(unitcube, phi, theta):
    chrgs, hier = unitcube
    p, unqs, invs = hier["p"], hier["unqs"], hier["invs"]
    exp_pos = jnp.array([-1.0,2.3,-1.12])

    # polar matrix coeffs
    basemat = get_polar_rot_coeff(p,unqs,jnp.array([[theta]]))

    # initial mpls
    locs_og = dbg.get_locs(hier["pts"][None,...], chrgs[None,...],exp_pos[None,...], p)

    locs = rot_azimuth(locs_og,jnp.array([[phi]]),False,True)
    locs = rot_polar(basemat,invs,locs,False,True)
    locs_back = rot_polar(basemat,invs,locs,True,True)
    locs_back = rot_azimuth(locs_back,jnp.array([[phi]]),True,True)

    spacerot = Rotation.from_euler('zy', jnp.array([-phi,-theta]))
    locs_manualrot = dbg.get_locs(spacerot.apply(hier["pts"])[None,...], chrgs[None,...],spacerot.apply(exp_pos)[None,:],p)

    assert jnp.allclose(locs_og,locs_back,rtol=1e-2,atol=1e-2)      # test inverse rotation
    assert jnp.allclose(locs_manualrot,locs,rtol=1e-2,atol=1e-2)    # test forward rotation


import jax.numpy as jnp
from jax.scipy.optimize import minimize
from jaxfmm import *

print("Assembling grid of charges.")
nside, sidelen = 15, 4*jnp.pi
pts = (jnp.mgrid[:nside,:nside,:nside].T/(nside-1) * sidelen - sidelen/2).reshape((-1,3))
chrgs = 0.01*jnp.ones(pts.shape[0])

print("Generating hierarchy.")
tree_info = gen_hierarchy(pts)

print("Computing desired potential.")
desired_pot = jnp.sin(jnp.linalg.norm(pts,axis=-1))
norm = jnp.linalg.norm(desired_pot)

def loss(chrgs):
    return jnp.linalg.norm(desired_pot-eval_potential(chrgs, **tree_info))/norm

print("Initial loss: %.2e"%loss(chrgs))
print("Compiling + running minimizer.")
res = minimize(loss,chrgs,method="BFGS",options={"maxiter": 1000})
print("Final loss (%i iterations): %.2e"%(res.nit,res.fun))


import jax.numpy as jnp
import jaxfmm.hierarchy as trees

### TODO: find better tests for the hierarchy gen

def test_tree_connectivity_unitcube():
    nside, sidelen = 64, 10.0
    pts = (jnp.mgrid[:nside,:nside,:nside].T/(nside-1) * sidelen - sidelen/2).reshape((-1,3))

    max_l = trees.get_max_l(pts.shape[0], 128)
    assert max_l==4

    idcs, rev_idcs, boxcenters, boxlens = trees.balanced_tree(pts,max_l)
    assert idcs.shape[1]==64
    assert jnp.all(idcs < pts.shape[0])
    assert jnp.allclose(pts, pts[idcs].reshape((-1,3))[rev_idcs])
    for i in range(len(boxlens)):
        assert jnp.allclose(boxlens[i], pts[nside//(2**i)-1,0] - pts[0,0])

    mpl_cnct, dir_cnct, _, _ = trees.gen_connectivity(boxcenters, boxlens,boxcenters,boxlens,no_cross_level=True)

    assert (mpl_cnct[-1]<mpl_cnct[-1].shape[0]).sum() == 611136
    assert (dir_cnct<dir_cnct.shape[0]).sum() == 70336